#!/usr/bin/perl

use v5.10;
use strict;
use warnings;

my $version = "1.0";

# Internal keys to MC field name mappings.  
# The names in the MC Field column should match the MC field names
# you want to use.  Customize the field name mappings to match your
# field names where necessary.
my %MCFieldMappings = (
  # key              MC Field           parse function
    url		=> [ 'AMG URL',		undef ],
    rating	=> [ 'AMG Rating',	\&get_rating ],
    pick	=> [ 'AMG Track Pick',	undef ],
    review	=> [ 'AMG Review',	\&get_review ],
    styles	=> [ 'Styles',		\&get_styles ],		# list
    moods	=> [ 'Moods',		\&get_moods ],		# list
    themes	=> [ 'Themes',		\&get_themes ],		# list

    # These are standard MC fields.
    name	=> [ 'Name',		undef ],
    composer	=> [ 'Composer',	undef ],
    artist	=> [ 'Artist',		undef ],		# list
    credits	=> [ 'People',		undef ],		# list
    date	=> [ 'Date',		\&get_releasedate ],
    album	=> [ 'Album',		\&get_album ],
    genre	=> [ 'Genre',		\&get_genre ],
    # The raw AMG rating is mapped above. This version is scaled to fit MC's 5-Star values
    ratingmc	=> [ 'Rating',		\&get_rating_mc ],
    tracknum	=> [ 'Track #',		undef ]			# don't rename this one
);

require LWP::UserAgent;
use Getopt::Std;
use File::Basename;
use Date::Calc qw(Delta_Days);
use Time::HiRes qw(usleep);

my $sleep_microseconds = 500;
sub debug;
sub print_mpl;

my $prog = basename($0);
my %opts;

# d=debug
# v=verbose
# f=local html file (debug)
my $opt_letters = "dvf:u:";
getopts($opt_letters, \%opts) or
    die "Usage: $prog [-$opt_letters]\nStopped";

# Load the OS-specific clipboard module
BEGIN {
    my ($found, $mod) = (0,0);
    for $mod ( qw(Win32::Clipboard Mac::Pasteboard) ) {
	if (eval "require $mod") {
	    $mod->import();
	    $found = 1;
	    last;
	}
    }
    die "Unable to load clipboard module\nStopped" unless $found;
}


my ($URL, @cliptracks, @clipheaders, %header2index, $amgcredits);
my %values;
my $clip = undef;
my @cliplines;

if ($^O eq 'MSWin32') {
    $clip = Win32::Clipboard();
    @cliplines = split /\r\n/, $clip->GetText();
}
else {
    $clip = Mac::Pasteboard->new ();
    @cliplines = split /\r\r/, $clip->paste();
}

@clipheaders = split /\t/, shift @cliplines;
push @cliptracks, [ split(/\t/, $_) ]  for @cliplines;

for (my $i = 0; $i < @clipheaders; $i++) {
    #say "CLIPHEADERS: $i: ", $clipheaders[$i];
    $header2index{$clipheaders[$i]} = $i;
}

defined($header2index{'Track #'}) or
    die "Field 'Track #' column is missing from the clipboard.  Add 'Track #' field to the view\nStopped";

if ($opts{'f'}) {
    # debugging
    local $/ = undef;
    open(FILE, "<", $opts{'f'}) or die "can't open file $opts{'f'}\nStopped";
    $_ = <FILE>;
    close FILE;
}
else {
    my $URL;

    if ($opts{'u'}) {
	$URL = $opts{'u'};
    }
    elsif (defined $header2index{mcfieldname('url')}) {
	($URL = $cliptracks[0][$header2index{mcfieldname('url')}]) or
	    die 'The field ', mcfieldname('url'), " is empty.  Fill the field with a valid AMG URL or use -u\nStopped";
    }
    else {
	die "The field ", mcfieldname('url'), " is missing from the clipboard.  Please add your AMG URL column and re-copy.\nStopped";
    }

    $values{'url'} = $URL;

    my $ua = LWP::UserAgent->new(agent => 'Mozilla/5.0');
    $ua->timeout(10);
    $ua->env_proxy;

    debug "Getting page from ", $URL;
    my $response = $ua->get($URL);
    if ($response->is_success) {
	$_ = $response->content;
    }
    else {
	die "Failed to get page at URL: $URL\n", $response->status_line, "\nStopped";
    }

    # don't bang on the server too fast
    usleep ($sleep_microseconds);

    $URL = join('/', $URL, 'credits');
    $response = $ua->get($URL);
    if ($response->is_success) {
	local $_ = $response->content;

	# AMG Credits page parsing
	m#<tbody>(.*?)</tbody>#ms or
	    die "Failed to parse credits\nStopped";

	$_ = $1;

	my ($s, @credits);
	while (s%<td class="artist".*?}">(.*?)</a>.*?<td class="credit"[^>]+>\s*(.*?)\s*</td>%%ms) {
	    my ($name, $credit) = ($1, $2);
	    $credit =~ s/,  /, /g;
	    push @credits, "$name [$credit]";
	}
	$values{'amgcredits'} = join(';', @credits);
    }
    else {
	die "Failed to get credits page at URL: $URL\n", $response->status_line, "\nStopped";
    }
}

# a few cleanups to make parsing easier
s/[ \t]+/ /g;
s#<a href[^>]+>#<A>#g;

# the track listing is a big chunk of HTML - pull it out early
s#\s*<section class="track-listing">.+?<tbody[^>]*>(.+?)</tbody>##sm;
my $tracklisting = $1;
$tracklisting =~ s/\n+/\n/gsm;

# general parser dispatch loop
for my $fields (keys %MCFieldMappings) {
    if ($MCFieldMappings{$fields}->[1]) {
	$values{$fields} = $MCFieldMappings{$fields}->[1]($_);
    }
}

# fieldkeys retains the key order to maintain the field name / field value relationship
my @fieldkeys = keys %MCFieldMappings;
my @amgtracks = ();
push @amgtracks, [ mcfieldname(@fieldkeys) ];

while (
    $tracklisting =~ s%
      <tr\ class="track(?<pick>\ pick)?
      .*?<td\ class="tracknum">\s*(?<tracknum>\d+)
      .*?<div\ class="title".*?<A>(?<name>.+?)</a>
      .*?<div\ class="composer">\s*<A>(?<composer>.*?)\s*</div>
      .*?<div\ class="primary".*?<A>(?<artist>.*?)\s*</div>
     %%xms)
{
    $values{'tracknum'}	= $+{'tracknum'};
    $values{'name'}	= clean($+{'name'});
    $values{'composer'}	= clean($+{'composer'});
    $values{'artist'}	= clean($+{'artist'});
    $values{'pick'}	= defined $+{'pick'} ? 1 : 0;

    for (@fieldkeys) {
	push @{$amgtracks[$+{'tracknum'}]}, $values{$_}
    }
}

die "AMG HTML has changed.  $0 needs to be updated for the new format.\nStopped" if @amgtracks == 1;

if ($clip) {
    print_mpl '<?xml version="1.0" encoding="UTF-8" standalone="yes" ?>';
    print_mpl '<MPL Version="2.0" Title="Created from ', $0, '">';

    #local $, = "\t";
    #say $track, clean($name), clean($composer), clean($performer), $featuring ? clean($featuring) : '';

    # add Filename field to headers
    push @{$amgtracks[0]}, 'Filename';
    for (@cliptracks) {
	my $i = $_->[$header2index{'Track #'}];
	#print "Track: $i, $_->[$header2index{'Filename'}], $amgtracks[$i][0]\n";
	push @{$amgtracks[$i]}, $_->[$header2index{'Filename'}];
	print_item($amgtracks[$i], $amgtracks[0]);
    }

    print_mpl '</MPL>';
}

# specific component parser functions
sub get_releasedate {
    local $_ = shift;
    my $ret;

    m#<div class="release-date">\s*<h4>Release Date</h4>\s*<span>(.+?)</span># or
	die "Release date not matched\nStopped";

    my $releasedate = $1;
    if ($releasedate =~ /\s*(\d{4})$/) {
	my @MCepoch = (1899, 12, 31);
	$ret = 1 + Delta_Days(@MCepoch, ($1,1,1));
    }
    debug "$ret ($releasedate)";
    return $ret;
}

sub get_recorddate {
    local $_ = shift;

    m#<div class="recording-date">\s*<h4>Recording Date</h4>\s*<div>(.+?)</div># or
	die "Record date not matched\nStopped";

    debug $1;
    return $1;
}

sub get_album {
    local $_ = shift;

    /<h2 class="album-title[^>]+>\s*(.*?)\s*<\/h2>/ or
	die "Album not matched\nStopped";

    debug $1;
    return $1;
}

sub get_rating {
    local $_ = shift;

    /<div class="allmusic-rating rating-allmusic-(\d)"/ or
	die "Rating not matched\nStopped";

    debug $1;
    return $1;
}

sub get_rating_mc {
    my $_ = shift;

    my $r = get_rating($_);
    return $r == 0 ? 0 : sprintf("%d", ($r / 2) + .5);
}

sub get_genre {
    local $_ = shift;

    m#<div class="genre">\s*<h4>Genre</h4>\s*<div>\s*<A>(.+?)</a>\s*</div># or
	die "Genre not matched\nStopped";

    debug $1;
    return $1;
}

sub get_styles {
    local $_ = shift;

    m#<div class="styles">\s*<h4>Styles</h4>\s*<div>\s*<A>(.+?)</a>\s*</div># or
	die "Styles not matched\nStopped";

    my $ret = $1;
    $ret =~ s#</a> <A>#;#g;
    $ret = clean($ret);
    debug $ret;
    return $ret;
}

sub get_moods {
    local $_ = shift;

    m#\s*<section\s+class="moods">.+?<div>(.*?)</div>#sm or
	die "Moods not matched\nStopped";

    my $ret = $1;
    $ret =~ s#<span class="mood">\s*<A>(.+?)</a>\s*</span>\s*#$1;#gsm;
    $ret =~ s/;$//;
    $ret = clean($ret);
    debug $ret;
    return $ret;
}

sub get_themes {
    local $_ = shift;

    m#\s*<section\s+class="themes">.+?<div>(.*?)</div>#sm or
	die "Themes not matched\nStopped";

    my $ret = $1;
    $ret =~ s#<span class="theme">\s*<A>(.+?)</a>\s*</span>\s*#$1;#gsm;
    $ret =~ s/;$//;
    $ret = clean($ret);
    debug $ret;
    return $ret;
}

sub get_review {
    local $_ = shift;

    m#<h4 class="review-author headline">\s*Review\s+by <span [^>]+>(.+?)</span>\s*</h4>\s*<div class="text"[^>]+>\s*<p>\s*(.+?)\s*</p>\s*</div>#ms or
	die "Review not matched\nStopped";

    my $ret = clean($2) .  " [$1]";
    debug $ret;
    return $ret;
}


sub mcfieldname {
    return wantarray ? map($MCFieldMappings{$_}[0], @_) : $MCFieldMappings{$_[0]}[0];
}

sub clean {
    local $_ = shift;
    s/[\n\r]//g;
    s/<A>//g;
    s/<\/a>(\s*<\/span>)?//g;
    s/feat. //;
    s/ \/ /;/g;
    s/^\s+//;
    s/\s+$//;
    return $_;
}

sub print_mpl {
    local $\ = "\n";

    print @_;
}

sub print_item {
    my ($cellsref, $namesref) = @_;

    print_mpl '<Item>';
    for (my $i = 0; $i < @$namesref; $i++) {
	unless (!$cellsref->[$i] or $cellsref->[$i] eq '' or $cellsref->[$i] eq '-') {
	    $cellsref->[$i] =~ s/\&/&amp;/g;
	    $cellsref->[$i] =~ s/</&lt;/g;
	    $cellsref->[$i] =~ s/>/&gt;/g;
	    print_mpl '<Field Name="', $namesref->[$i], '">', $cellsref->[$i], '</Field>'
	}
    }
    print_mpl '<Field Name="Media Type">Audio</Field>';
    print_mpl '</Item>';
}

sub debug {
    return unless $opts{'d'};

    say((split(/::/, (caller(1))[3] || 'main'))[-1], ': ', @_);
}
